part of masamune_builder;

Builder masamuneBuilderFactory(BuilderOptions options) {
  return PartBuilder([MasamuneCollectionGenerator()], ".m.dart");
}

class MasamuneCollectionGenerator
    extends GeneratorForAnnotation<CollectionPath> {
  @override
  FutureOr<String> generateForAnnotatedElement(
    Element element,
    ConstantReader annotation,
    BuildStep buildStep,
  ) {
    if (!element.library!.isNonNullableByDefault) {
      throw InvalidGenerationSourceError(
        "Generator cannot target libraries that have not been migrated to "
        "null-safety.",
        element: element,
      );
    }

    if (element is! ClassElement) {
      throw InvalidGenerationSourceError(
        "`@CollectionPath()` can only be used on classes.",
        element: element,
      );
    }

    final path = annotation.read("path").stringValue;
    final name = element.displayName;
    final constructor = element.constructors;
    final params = element.constructors.first.parameters;

    print(name);
    print(constructor);
    constructor.forEach((e) => print(e.name));
    print(params);

    if (path.splitLength() <= 0 || path.splitLength() % 2 != 1) {
      throw Exception(
        "The path hierarchy must be an odd number: $path",
      );
    }

    return "";

    // final mixin = Library(
    //   (l) => l
    //     ..body.addAll(
    //       [
    //         Class(
    //           (c) => c..name = "",
    //         )
    //         Method(
    //           (m) => m
    //             ..name = "_\$${element.name}FromMap"
    //             ..requiredParameters.addAll([
    //               Parameter(
    //                 (b) => b
    //                   ..name = "map"
    //                   ..type = const Reference("DynamicMap"),
    //               ),
    //               Parameter(
    //                 (b) => b
    //                   ..name = "ref"
    //                   ..type = Reference(element.name),
    //               )
    //             ])
    //             ..returns = refer("${element.name}?")
    //             ..body = Code(
    //                 "if(map.isEmpty || map.get(\"type\", \"\") != ref.type){ return null; } return ${element.name}(${_generateFromMapCode(element)});"),
    //         ),
    //         Method(
    //           (m) => m
    //             ..name = "_\$${element.name}ToMap"
    //             ..requiredParameters.addAll([
    //               Parameter(
    //                 (b) => b
    //                   ..name = "ref"
    //                   ..type = Reference(element.name),
    //               )
    //             ])
    //             ..returns =
    //                 refer("DynamicMap", "package:masamune/masamune.dart")
    //             ..body = Code(
    //                 "return <String, dynamic>{${_generateToMapCode(element)}};"),
    //         ),
    //       ],
    //     ),
    // );
    // final emitter = DartEmitter();
    // return DartFormatter().format("${mixin.accept(emitter)}");
  }

  String _generateFromMapCode(ClassElement element) {
    return [
      ...element.constructors.first.parameters
          .map(
            (e) => e.codeFromMap,
          )
          .where((e) => e.isNotEmpty),
    ].join(",");
  }

  String _generateToMapCode(ClassElement element) {
    return [
      "\"type\": ref.type",
      ...element.constructors.first.parameters
          .map(
            (e) => e.codeToMap,
          )
          .where((e) => e.isNotEmpty),
    ].join(",");
  }
}

class ModelClass {
  ModelClass(this.element) {
    name = element.displayName;
    final contstuctor = element.constructors.firstWhere((e) {
      return e.isFactory && e.name.isEmpty;
    });
    parameters = contstuctor.parameters.map((e) {
      return ModelParam(e);
    }).toList();
  }

  final ClassElement element;

  late final String name;
  late final List<ModelParam> parameters;
}

class ModelParam {
  ModelParam(this.element) {
    name = element.displayName;
    type = element.type;
    if (element.type.isNullable || element.isRequired) {
      defaultValue = null;
    } else {
      element.metadata
          .firstWhereOrNull((item) => item.element?.displayName == "Default");
    }
  }
  final ParameterElement element;
  late final Object? defaultValue;
  late final DartType type;
  late final String name;
}
